# encoding: utf8
"""
query.py
@author Meng.yangyang
@description 信息查询
@created Mon Jan 07 2019 16:50:59 GMT+0800 (CST)
"""

import re
import copy
import json
import logging

from hack12306.constants import SEAT_TYPE_CODE_MAP
from hack12306.query import TrainInfoQueryAPI
from hack12306.constants import SEAT_TYPE_CODE_MAP

from . import exceptions

_logger = logging.getLogger('booking')

__all__ = ('query_left_tickets', 'query_station_code_map',)


def _check_seat_type_is_booking(left_ticket):
    if left_ticket and left_ticket != u'无' and left_ticket != u'*':
        return True
    else:
        return False


def _select_train_and_seat_type(train_names, seat_types, query_trains):
    """
    选择订票车次、席别
    :param train_names 预定的车次列表
    :param seat_types 预定席别列表
    :param query_trains 查询到火车车次列表
    :return select_train, select_seat_type
    """
    def _select_trains(query_trains, train_names=None):
        if train_names:
            select_trains = []
            # 根据订票车次次序，选择车次
            for train_name in train_names:
                for train in query_trains:
                    if train['train_name'] == train_name:
                        select_trains.append(copy.deepcopy(train))
            return select_trains
        else:
            return query_trains

    def _select_types(trains, seat_types):
        select_train = None
        select_seat_type = None

        for train in trains:
            for seat_type in seat_types:
                seat_type_left_ticket = train.get(seat_type, '')
                if _check_seat_type_is_booking(seat_type_left_ticket):
                    select_seat_type = seat_type
                    select_train = copy.deepcopy(train)
                    return select_train, select_seat_type
        else:
            return None, None

    _logger.debug('train_names:%s seat_types:%s' % (json.dumps(train_names, ensure_ascii=False),
                                                    json.dumps(seat_types, ensure_ascii=False)))
    trains = _select_trains(query_trains, train_names)
    # debug trains
    for i in range(min(len(trains), len(train_names or ['']))):
        _logger.debug('query left tickets train info. %s' % json.dumps(trains[i], ensure_ascii=False))

    return _select_types(trains, seat_types)


def query_left_tickets(train_date, from_station, to_station, seat_types, train_names=None):
    """
    信息查询-剩余车票
    :param train_date
    :param from_station
    :param to_station
    :param seat_types
    :param train_names
    :return JSON 对象
    """
    date_pattern = re.compile(r'^\d{4}-\d{2}-\d{2}$')
    assert date_pattern.match(train_date), 'Invalid train_date param. %s' % train_date

    assert isinstance(seat_types, list), u'Invalid seat_types param. %s' % seat_types
    assert frozenset(seat_types) <= frozenset(dict(SEAT_TYPE_CODE_MAP).keys()
                                              ), u'Invalid seat_types param. %s' % seat_types

    train_info = {}

    trains = TrainInfoQueryAPI().info_query_left_tickets(train_date, from_station, to_station)
    train_info, select_seat_type = _select_train_and_seat_type(train_names, seat_types, trains)

    if not train_info or not select_seat_type:
        raise exceptions.BookingTrainNoLeftTicket()

    _logger.debug('select train info. %s' % json.dumps(train_info, ensure_ascii=False))

    result = {
        'train_date': train_date,
        'from_station': train_info['from_station'],
        'to_station': train_info['to_station'],
        'seat_type': select_seat_type,
        'seat_type_code': dict(SEAT_TYPE_CODE_MAP)[select_seat_type],
        'departure_time': train_info['departure_time'],
        'arrival_time': train_info['arrival_time'],
        'secret': train_info['secret'],
        'train_name': train_info['train_name'],
        'duration': train_info['duration'],
        'train_num': train_info['train_num']
    }
    return result


def query_station_code_map():
    """
    信息查询-查询车站编码列表
    :return JSON对象
    """
    station_code_map = {}

    stations = TrainInfoQueryAPI().info_query_station_list()
    for station in stations:
        station_code_map[station['name']] = station['code']

    return station_code_map



# encoding: utf8
"""
auth.py
@author Meng.yangyang
@description 认证
@created Mon Jan 07 2019 16:35:01 GMT+0800 (CST)
"""


import os
import time
import uuid
import json
import base64
import logging

from hack12306.auth import TrainAuthAPI
from hack12306.user import TrainUserAPI
from hack12306.exceptions import TrainUserNotLogin

from . import settings

_logger = logging.getLogger('booking')

__all__ = ('auth_qr', 'auth_reauth', 'auth_is_login')


def _uamtk_set(uamtk):
    settings.AUTH_UAMTK = uamtk


def _uamtk_get():
    return settings.AUTH_UAMTK


def auth_is_login(cookies=None):
    """
    检查用户是否登录
    :param cookies JSON对象
    :return True已登录, False未登录
    """
    result = TrainAuthAPI().auth_check_login(cookies=cookies)
    if not result:
        _logger.debug('会话已过期，请重新登录!')
    return result


def auth_reauth(uamtk, cookie_dict):
    """
    重新认证
    :param aumtk
    :param cookie_dict
    :return JSON对象
    """
    assert uamtk is not None
    assert isinstance(cookie_dict, dict)

    train_auth_api = TrainAuthAPI()

    uamtk_result = train_auth_api.auth_uamtk(uamtk, cookies=cookie_dict)
    _logger.debug('4. auth uamtk result. %s' % json.dumps(uamtk_result, ensure_ascii=False))

    uamauth_result = train_auth_api.auth_uamauth(uamtk_result['newapptk'], cookies=cookie_dict)
    _logger.debug('5. auth uamauth result. %s' % json.dumps(uamauth_result, ensure_ascii=False))

    return uamauth_result


def auth_qr():
    """
    认证-二维码登录
    """
    try:
        qr_img_path = '/tmp/12306/booking/login-qr-%s.jpeg' % uuid.uuid1().hex

        train_auth_api = TrainAuthAPI()

        _logger.debug('1. auth init')
        cookie_dict = train_auth_api.auth_init()

        _logger.debug('2. auth get qr')
        result = train_auth_api.auth_qr_get(cookies=cookie_dict)
        assert isinstance(result, dict)
        qr_uuid = result['uuid']

        if not os.path.exists(os.path.dirname(qr_img_path)):
            os.makedirs(os.path.dirname(qr_img_path))

        with open(qr_img_path, 'wb') as f:
            f.write(base64.b64decode(result['image']))

        # open qr image with browser
        cmd = settings.CHROME_APP_OPEN_CMD.format(filepath=qr_img_path)
        os.system(cmd)

        _logger.debug('3. auth check qr')
        for _ in range(15):
            _logger.info('请扫描二维码登录！')
            qr_check_result = train_auth_api.auth_qr_check(qr_uuid, cookies=cookie_dict)
            _logger.debug('check qr result. %s' % json.dumps(qr_check_result, ensure_ascii=False))
            if qr_check_result['result_code'] == "2":
                _logger.debug('qr check success result. %s' % json.dumps(qr_check_result, ensure_ascii=False))
                _logger.info('二维码扫描成功！')
                break

            time.sleep(1)
        else:
            _logger.error('二维码扫描失败，重新生成二维码')
            raise TrainUserNotLogin('扫描述二维码失败')

        _uamtk_set(qr_check_result['uamtk'])
        uamauth_result = auth_reauth(_uamtk_get(), cookie_dict)
        _logger.info('%s 登录成功。' % uamauth_result['username'].encode('utf8'))

        cookies = {
            'tk': uamauth_result['apptk']
        }
        cookies.update(**cookie_dict)
        _logger.debug('cookies. %s' % json.dumps(cookies, ensure_ascii=False,))

        # user_info_result = TrainUserAPI().user_info(cookies=cookies)
        # _logger.debug('%s login successfully.' % user_info_result['name'])

        return cookies
    finally:
        if os.path.exists(qr_img_path):
            os.remove(qr_img_path)



# encoding: utf8
"""
utils.py
@author Meng.yangyang
@description 工具函数
@created Mon Jan 07 2019 13:22:25 GMT+0800 (CST)
"""

import os
import json
import platform
import requests


# def qr_terminal_draw(filepath):
#     from PIL import Image
#     assert isinstance(filepath, (str))

#     if not os.path.exists(filepath):
#         raise Exception('file not exists. %s' % filepath)

#     if platform.system() == "Windows":
#         white_block = '▇'
#         black_block = '  '
#         new_line = '\n'
#     else:
#         white_block = '\033[1;37;47m  '
#         black_block = '\033[1;37;40m  '
#         new_line = '\033[0m\n'

#     output = ''
#     im = Image.open(filepath)
#     im = im.resize((21, 21))

#     pixels = im.load()

#     output += white_block * (im.width + 2) + new_line
#     for h in range(im.height):
#         output += white_block
#         for w in range(im.width):
#             pixel = pixels[w,h]     # NOQA
#             if pixel[0] == 0:
#                 output += black_block
#             elif pixel[0] == 255:
#                 output += white_block
#             else:
#                 assert 'Unsupported pixel. %s' % pixel

#         else:
#             output += white_block + new_line
#     output += white_block * (im.width + 2) + new_line

#     return output


def get_public_ip():
    resp = requests.get('http://httpbin.org/ip')
    if resp.status_code != 200:
        raise Exception('Network error')
    return json.loads(resp.content)['origin'].encode('utf8')


def check_seat_types(seat_types):
    from hack12306.constants import SEAT_TYPE_CODE_MAP

    assert isinstance(seat_types, (list, tuple))
    if not frozenset(seat_types) <= frozenset(dict(SEAT_TYPE_CODE_MAP).keys()):
        return False
    return True



# encoding: utf8
"""
pay.py
@author Meng.yangyang
@description 支付
@created Mon Jan 07 2019 17:33:55 GMT+0800 (CST)
"""

import os
import json
import copy
import logging
import datetime
import platform

from hack12306 import constants
from hack12306.pay import TrainPayAPI
from hack12306.utils import tomorrow, JSONEncoder

from . import settings
from . import exceptions
from .order import order_no_complete
from .utils import get_public_ip

_logger = logging.getLogger('booking')

__all__ = ('pay_order', )


def pay_order(bank_id=constants.BANK_ID_WX, **kwargs):
    """
    支付订单
    :param sequence_no 订单后
    :bank_id 支付渠道ID
    :return None
    """
    train_pay_api = TrainPayAPI()

    # 0.查询未支付订单
    sequence_no = order_no_complete()
    if not sequence_no:
        raise exceptions.BookingOrderNoExists('')

    # 1.支付未完成订单
    pay_no_complete_order_result = train_pay_api.pay_no_complete_order(sequence_no, cookies=settings.COOKIES)
    _logger.debug('pay no complete order result. %s' % json.dumps(pay_no_complete_order_result, ensure_ascii=False,))
    if pay_no_complete_order_result['existError'] != 'N':
        raise exceptions.BookingOrderNoExists('%s订单不存在' % sequence_no)

    # 2.支付初始化
    train_pay_api.pay_init(cookies=settings.COOKIES)

    # 3.发起支付
    pay_check_new_result = train_pay_api.pay_check_new(cookies=settings.COOKIES)
    _logger.debug('pay check new result. %s' % json.dumps(pay_check_new_result, ensure_ascii=False))

    # 4.交易
    pay_business_result = train_pay_api.pay_web_business(
        pay_check_new_result['payForm']['tranData'],
        pay_check_new_result['payForm']['merSignMsg'],
        pay_check_new_result['payForm']['transType'],
        get_public_ip(), pay_check_new_result['payForm']['tranDataParsed']['order_timeout_date'],
        bank_id, cookies=settings.COOKIES)
    _logger.debug('pay business result. %s' % json.dumps(pay_business_result, ensure_ascii=False))

    # 5.跳转第三方支付
    pay_business_third_pay_resp = train_pay_api.submit(
        pay_business_result['url'],
        pay_business_result['params'],
        method=pay_business_result['method'],
        parse_resp=False,
        cookies=settings.COOKIES,
        allow_redirects=True)
    _logger.debug('pay third resp status code. %s' % pay_business_third_pay_resp.status_code)
    _logger.debug('pay third resp. %s' % pay_business_third_pay_resp.content)

    # 6.打开浏览器扫码完成支付
    try:
        pay_filepath = settings.PAY_FILEPATH.format(date=datetime.date.today().strftime('%Y%m%d'),
                                                    order_no=sequence_no, bank_id=bank_id)
        if not os.path.exists(os.path.dirname(pay_filepath)):
            os.makedirs(os.path.dirname(pay_filepath))

        with open(pay_filepath, 'w') as f:
            f.write(pay_business_third_pay_resp.content)

        _logger.info('请用浏览器打开%s，完成支付！' % pay_filepath)
    finally:
        if os.path.exists(pay_filepath):
            if platform.mac_ver()[0]:
                os.system('open %s' % pay_filepath)

            os.remove(pay_filepath)



# encoding: utf8

import os
import logging

INIT_DONE = False

# Auth settings
AUTH_UAMTK = None
AUTH_REAUTH_INTERVAL = 60 * 3  # 单位：秒

STATION_LIST_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'station_list.json')
QUERY_LEFT_TICKET_COUNTER_FILE = '/tmp/12306-booking/left_ticket_counter'
SLEEP_INTERVAL = 0.6

COOKIES = {}
PAY_FILEPATH = './{date}-{order_no}-{bank_id}.html'
STATION_CODE_MAP = {}

CHROME_APP_OPEN_CMD_MacOS = 'open -a /Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome {filepath}'
CHROME_APP_OPEN_CMD_LINUX = None    # TODO Linux Chrome 打开文件cmd
CHROME_APP_OPEN_CMD_WINDOWS = None  # TODO Windows Chrome 打开文件cmd
CHROME_APP_OPEN_CMD = CHROME_APP_OPEN_CMD_MacOS

LOGGING = {
    'version': 1,
    'formatters': {
        'default': {
            'format': '%(asctime)s - %(levelname)s - %(message)s'
        },
        'app': {
            'format': '%(asctime)s - %(levelname)s - %(module)s::%(funcName)s:%(lineno)d - %(message)s'
        },
    },
    'filters': {
        'log_level': {
            '()': 'booking._logging.LogLevelFilter',
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'default',
            'level': 'INFO',
            'filters': ['log_level'],
        },
        'app': {
            'class': 'logging.handlers.WatchedFileHandler',
            'filename': 'app.log',
            'formatter': 'app',
            'level': 'DEBUG',
        }
    },
    'loggers': {
        'booking': {
            'handlers': ['console', 'app'],
            'level': os.getenv('BOOKING_LOG_LEVEL', 'INFO'),
        }
    },
}




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="12306-booking",
    version="0.1.11",
    author="Meng.yangyang",
    author_email="mengyy_linux@163.com",
    description="12306 booking tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    url="https://github.com/hack12306/12306-booking",
    packages=setuptools.find_packages(),
    include_package_data = True,
    package_data = {'': 'booking/station_list.json'},
    # install_requires=["Pillow>=5.4.1", "hack12306>=0.1.11", "click==7.0", "six>=1.12.0"],
    install_requires=["hack12306>=0.1.11", "click==7.0", "six>=1.12.0"],
    entry_points={
        'console_scripts': [
            '12306-booking=booking.command:booking'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 2",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



# encoding: utf8
"""
user.py
@author Meng.yangyang
@description User
@created Wed Jan 09 2019 21:17:23 GMT+0800 (CST)
"""

import json
import logging
from hack12306.user import TrainUserAPI

from . import settings

__all__ = ('user_passengers',)

_logger = logging.getLogger('booking')


def user_passengers():
    """
    User passengers
    """
    passengers = TrainUserAPI().user_passengers(cookies=settings.COOKIES)
    _logger.debug(json.dumps(passengers, ensure_ascii=False))
    return passengers



# encoding: utf8
"""
run.py
@author Meng.yangyang
@description Booking entry point
@created Tue Jan 08 2019 19:38:32 GMT+0800 (CST)
"""

import os
import re
import time
import json
import fcntl
import logging
import platform
import logging.config
from hack12306.constants import (BANK_ID_WX, BANK_ID_MAP, SEAT_TYPE_CODE_MAP,)
from hack12306.exceptions import TrainUserNotLogin, TrainBaseException

from . import settings
from . import exceptions
from .pay import pay_order
from .user import user_passengers
from .auth import auth_qr, auth_is_login, auth_reauth
from .order import order_submit, order_check_no_complete
from .query import query_left_tickets, query_station_code_map

_logger = logging.getLogger('booking')

__all__ = ('initialize', 'run')


BOOKING_STATUS_QUERY_LEFT_TICKET = 2
BOOKING_STATUS_ORDER_SUBMIT = 3
BOOKING_STATUS_PAY_ORDER = 4

BOOKING_STATUS_MAP = [
    (BOOKING_STATUS_QUERY_LEFT_TICKET, '查询余票'),
    (BOOKING_STATUS_ORDER_SUBMIT, '提交订单'),
    (BOOKING_STATUS_PAY_ORDER, '支付订单'),
]


def initialize():
    """
    Initialization.
    """
    if settings.INIT_DONE:
        return

    station_list = []
    station_code_map = {}
    with open(settings.STATION_LIST_FILE) as f:
        station_list = json.loads(f.read())

    for station in station_list:
        station_code_map[station['name']] = station['code']

    settings.STATION_CODE_MAP = station_code_map
    del station_list

    logging.config.dictConfig(settings.LOGGING)

    if platform.system() == "Windows":
        settings.CHROME_APP_OPEN_CMD = settings.CHROME_APP_OPEN_CMD_WINDOWS
    elif platform.system() == 'Linux':
        settings.CHROME_APP_OPEN_CMD = settings.CHROME_APP_OPEN_CMD_LINUX
    elif platform.mac_ver()[0]:
        settings.CHROME_APP_OPEN_CMD = settings.CHROME_APP_OPEN_CMD_MacOS
    else:
        settings.CHROME_APP_OPEN_CMD = settings.CHROME_APP_OPEN_CMD_MacOS

    settings.INIT_DONE = True


def _query_left_ticket_counter_get():
    if not os.path.exists(settings.QUERY_LEFT_TICKET_COUNTER_FILE):
        return 0

    with open(settings.QUERY_LEFT_TICKET_COUNTER_FILE) as f:
        flag = fcntl.fcntl(f.fileno(), fcntl.F_GETFD)
        fcntl.fcntl(f, fcntl.F_SETFD, flag | os.O_NONBLOCK)
        flag = fcntl.fcntl(f, fcntl.F_GETFD)

        counter = f.read()
        return int(counter)


def _query_left_ticket_counter_inc():
    counter = _query_left_ticket_counter_get() + 1
    with open(settings.QUERY_LEFT_TICKET_COUNTER_FILE, 'w') as f:
        f.write(str(counter))


def run(train_date, train_names, seat_types, from_station, to_station, pay_channel=BANK_ID_WX, passengers=None, **kwargs):
    """
    Booking entry point.
    """
    initialize()
    assert settings.INIT_DONE is True, 'No Initialization'

    date_patten = re.compile(r'^\d{4}-\d{2}-\d{2}$')
    assert date_patten.match(train_date), 'Invalid train_date param. %s' % train_date

    assert isinstance(seat_types, (list, tuple)), u'Invalid seat_types param. %s' % seat_types
    assert frozenset(seat_types) <= frozenset(dict(SEAT_TYPE_CODE_MAP).keys()
                                              ), u'Invalid seat_types param. %s' % seat_types

    assert from_station in settings.STATION_CODE_MAP.values(), 'Invalid from_station param. %s' % from_station
    assert to_station in settings.STATION_CODE_MAP.values(), 'Invalid to_station param. %s' % to_station
    assert pay_channel in dict(BANK_ID_MAP).keys(), 'Invalid pay_channel param. %s' % pay_channel

    train_info = {}
    order_no = None
    check_passengers = False
    passenger_id_nos = []
    booking_status = BOOKING_STATUS_QUERY_LEFT_TICKET

    last_auth_time = int(time.time())

    while True:
        try:
            # auth
            if booking_status != BOOKING_STATUS_QUERY_LEFT_TICKET and(
                    not settings.COOKIES or not auth_is_login(settings.COOKIES)):
                cookies = auth_qr()
                settings.COOKIES = cookies

            # reauth
            if booking_status != BOOKING_STATUS_QUERY_LEFT_TICKET and settings.AUTH_UAMTK and settings.COOKIES:
                if int(time.time()) - last_auth_time >= settings.AUTH_REAUTH_INTERVAL:
                    uamauth_result = auth_reauth(settings.AUTH_UAMTK, settings.COOKIES)
                    settings.COOKIES.update(tk=uamauth_result['apptk'])
                    last_auth_time = int(time.time())
                    _logger.info('%s 重新认证成功' % uamauth_result['username'].encode('utf8'))

            # check passengers
            if booking_status != BOOKING_STATUS_QUERY_LEFT_TICKET and not check_passengers:
                passenger_infos = user_passengers()
                if passengers:
                    passenger_name_id_map = {}
                    for passenger_info in passenger_infos:
                        passenger_name_id_map[passenger_info['passenger_name']] = passenger_info['passenger_id_no']

                    assert frozenset(passengers) <= frozenset(passenger_name_id_map.keys()), u'无效的乘客. %s' % json.dumps(
                        list(frozenset(passengers) - frozenset(passenger_name_id_map.keys())), ensure_ascii=False)

                    for passenger in passengers:
                        _logger.info(u'订票乘客信息。姓名：%s 身份证号:%s' % (passenger, passenger_name_id_map[passenger]))
                        passenger_id_nos.append(passenger_name_id_map[passenger])
                else:
                    passenger_id_nos = [passenger_infos[0]['passenger_id_no']]
                    _logger.info(
                        u'订票乘客信息。姓名:%s 身份证号:%s' %
                        (passenger_infos[0]['passenger_name'],
                         passenger_info['passenger_id_no']))

                check_passengers = True

            # order not complete
            if booking_status != BOOKING_STATUS_QUERY_LEFT_TICKET and order_check_no_complete():
                booking_status = BOOKING_STATUS_PAY_ORDER

            _logger.debug('booking status. %s' % dict(BOOKING_STATUS_MAP).get(booking_status, '未知状态'))

            # query left tickets
            if booking_status == BOOKING_STATUS_QUERY_LEFT_TICKET:
                _query_left_ticket_counter_inc()
                _logger.info('查询余票, 已查询%s次!' % _query_left_ticket_counter_get())
                train_info = query_left_tickets(train_date, from_station, to_station, seat_types, train_names)
                booking_status = BOOKING_STATUS_ORDER_SUBMIT

            # subit order
            elif booking_status == BOOKING_STATUS_ORDER_SUBMIT:
                try:
                    _logger.info('提交订单')
                    order_no = order_submit(passenger_id_nos, **train_info)
                except (TrainBaseException, exceptions.BookingBaseException) as e:
                    _logger.info('提交订单失败')
                    booking_status = BOOKING_STATUS_QUERY_LEFT_TICKET
                    _logger.exception(e)
                    continue
                else:
                    # submit order successfully
                    if order_no:
                        _logger.info('提交订单成功')
                        booking_status = BOOKING_STATUS_PAY_ORDER

            # pay
            elif booking_status == BOOKING_STATUS_PAY_ORDER:
                _logger.info('支付订单')
                pay_order(pay_channel)
                # pay success and exit
                return
            else:
                assert 'Unkown booking status. %s' % booking_status

        except TrainUserNotLogin:
            _logger.warn('用户未登录，请重新扫码登录')
            continue

        except TrainBaseException as e:
            _logger.error(e)
            _logger.exception(e)

        except Exception as e:
            if isinstance(e, AssertionError):
                _logger.exception(e)
                _logger.error('系统内部运行异常，请重新执行程序！')
                os._exit(-1)
            elif isinstance(e, exceptions.BookingOrderCancelExceedLimit):
                _logger.exception(e)
                _logger.error('用户今日订单取消次数超限，请明天再重新抢票！')
                os._exit(-2)
            else:
                _logger.exception(e)

        time.sleep(settings.SLEEP_INTERVAL)
